# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import sqlalchemy

from opentelemetry import trace
from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor
from opentelemetry.test.test_base import TestBase

POSTGRES_CONFIG = {
    "host": "127.0.0.1",
    "port": int(os.getenv("TEST_POSTGRES_PORT", "5432")),
    "user": os.getenv("TEST_POSTGRES_USER", "testuser"),
    "password": os.getenv("TEST_POSTGRES_PASSWORD", "testpassword"),
    "dbname": os.getenv("TEST_POSTGRES_DB", "opentelemetry-tests"),
}


class SQLAlchemyInstrumentTestCase(TestBase):
    """TestCase that checks if the engine is properly traced
    when the `instrument()` method is used.
    """

    def setUp(self):
        super().setUp()
        # create a traced engine with the given arguments
        SQLAlchemyInstrumentor().instrument()
        dsn = (
            "postgresql://%(user)s:%(password)s@%(host)s:%(port)s/%(dbname)s"
            % POSTGRES_CONFIG
        )
        self.engine = sqlalchemy.create_engine(dsn)

        # prepare a connection
        self.conn = self.engine.connect()

    def tearDown(self):
        # clear the database and dispose the engine
        self.conn.close()
        self.engine.dispose()
        SQLAlchemyInstrumentor().uninstrument()
        super().tearDown()

    def test_engine_traced(self):
        # ensures that the engine is traced
        rows = self.conn.execute("SELECT").fetchall()
        self.assertEqual(len(rows), 1)

        spans = self.memory_exporter.get_finished_spans()
        # trace composition
        self.assertEqual(len(spans), 2)
        span = spans[1]
        # check subset of span fields
        self.assertEqual(span.name, "SELECT opentelemetry-tests")
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)
        self.assertGreater((span.end_time - span.start_time), 0)



import asyncio
import os

import asyncpg

from opentelemetry.instrumentation.asyncpg import AsyncPGInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase
from opentelemetry.trace import StatusCode

POSTGRES_HOST = os.getenv("POSTGRESQL_HOST", "localhost")
POSTGRES_PORT = int(os.getenv("POSTGRESQL_PORT", "5432"))
POSTGRES_DB_NAME = os.getenv("POSTGRESQL_DB_NAME", "opentelemetry-tests")
POSTGRES_PASSWORD = os.getenv("POSTGRESQL_PASSWORD", "testpassword")
POSTGRES_USER = os.getenv("POSTGRESQL_USER", "testuser")


def async_call(coro):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)


class TestFunctionalAsyncPG(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        AsyncPGInstrumentor().instrument(tracer_provider=self.tracer_provider)
        self._connection = async_call(
            asyncpg.connect(
                database=POSTGRES_DB_NAME,
                user=POSTGRES_USER,
                password=POSTGRES_PASSWORD,
                host=POSTGRES_HOST,
                port=POSTGRES_PORT,
            )
        )

    def tearDown(self):
        AsyncPGInstrumentor().uninstrument()
        super().tearDown()

    def check_span(self, span):
        self.assertEqual(
            span.attributes[SpanAttributes.DB_SYSTEM], "postgresql"
        )
        self.assertEqual(
            span.attributes[SpanAttributes.DB_NAME], POSTGRES_DB_NAME
        )
        self.assertEqual(
            span.attributes[SpanAttributes.DB_USER], POSTGRES_USER
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_NAME], POSTGRES_HOST
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_PORT], POSTGRES_PORT
        )

    def test_instrumented_execute_method_without_arguments(self, *_, **__):
        async_call(self._connection.execute("SELECT 42;"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        self.assertIs(StatusCode.UNSET, spans[0].status.status_code)
        self.check_span(spans[0])
        self.assertEqual(spans[0].name, "SELECT")
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT 42;"
        )

    def test_instrumented_fetch_method_without_arguments(self, *_, **__):
        async_call(self._connection.fetch("SELECT 42;"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        self.check_span(spans[0])
        self.assertEqual(spans[0].name, "SELECT")
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT 42;"
        )

    def test_instrumented_remove_comments(self, *_, **__):
        async_call(self._connection.fetch("/* leading comment */ SELECT 42;"))
        async_call(
            self._connection.fetch(
                "/* leading comment */ SELECT 42; /* trailing comment */"
            )
        )
        async_call(self._connection.fetch("SELECT 42; /* trailing comment */"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 3)
        self.check_span(spans[0])
        self.assertEqual(spans[0].name, "SELECT")
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT],
            "/* leading comment */ SELECT 42;",
        )
        self.check_span(spans[1])
        self.assertEqual(spans[1].name, "SELECT")
        self.assertEqual(
            spans[1].attributes[SpanAttributes.DB_STATEMENT],
            "/* leading comment */ SELECT 42; /* trailing comment */",
        )
        self.check_span(spans[2])
        self.assertEqual(spans[2].name, "SELECT")
        self.assertEqual(
            spans[2].attributes[SpanAttributes.DB_STATEMENT],
            "SELECT 42; /* trailing comment */",
        )

    def test_instrumented_transaction_method(self, *_, **__):
        async def _transaction_execute():
            async with self._connection.transaction():
                await self._connection.execute("SELECT 42;")

        async_call(_transaction_execute())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(3, len(spans))
        self.check_span(spans[0])
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "BEGIN;"
        )
        self.assertIs(StatusCode.UNSET, spans[0].status.status_code)

        self.check_span(spans[1])
        self.assertEqual(
            spans[1].attributes[SpanAttributes.DB_STATEMENT], "SELECT 42;"
        )
        self.assertIs(StatusCode.UNSET, spans[1].status.status_code)

        self.check_span(spans[2])
        self.assertEqual(
            spans[2].attributes[SpanAttributes.DB_STATEMENT], "COMMIT;"
        )
        self.assertIs(StatusCode.UNSET, spans[2].status.status_code)

    def test_instrumented_failed_transaction_method(self, *_, **__):
        async def _transaction_execute():
            async with self._connection.transaction():
                await self._connection.execute("SELECT 42::uuid;")

        with self.assertRaises(asyncpg.CannotCoerceError):
            async_call(_transaction_execute())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(3, len(spans))

        self.check_span(spans[0])
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "BEGIN;"
        )
        self.assertIs(StatusCode.UNSET, spans[0].status.status_code)

        self.check_span(spans[1])
        self.assertEqual(
            spans[1].attributes[SpanAttributes.DB_STATEMENT],
            "SELECT 42::uuid;",
        )
        self.assertEqual(StatusCode.ERROR, spans[1].status.status_code)

        self.check_span(spans[2])
        self.assertEqual(
            spans[2].attributes[SpanAttributes.DB_STATEMENT], "ROLLBACK;"
        )
        self.assertIs(StatusCode.UNSET, spans[2].status.status_code)

    def test_instrumented_method_doesnt_capture_parameters(self, *_, **__):
        async_call(self._connection.execute("SELECT $1;", "1"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        self.assertIs(StatusCode.UNSET, spans[0].status.status_code)
        self.check_span(spans[0])
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT $1;"
        )


class TestFunctionalAsyncPG_CaptureParameters(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        AsyncPGInstrumentor(capture_parameters=True).instrument(
            tracer_provider=self.tracer_provider
        )
        self._connection = async_call(
            asyncpg.connect(
                database=POSTGRES_DB_NAME,
                user=POSTGRES_USER,
                password=POSTGRES_PASSWORD,
                host=POSTGRES_HOST,
                port=POSTGRES_PORT,
            )
        )

    def tearDown(self):
        AsyncPGInstrumentor().uninstrument()
        super().tearDown()

    def check_span(self, span):
        self.assertEqual(
            span.attributes[SpanAttributes.DB_SYSTEM], "postgresql"
        )
        self.assertEqual(
            span.attributes[SpanAttributes.DB_NAME], POSTGRES_DB_NAME
        )
        self.assertEqual(
            span.attributes[SpanAttributes.DB_USER], POSTGRES_USER
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_NAME], POSTGRES_HOST
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_PORT], POSTGRES_PORT
        )

    def test_instrumented_execute_method_with_arguments(self, *_, **__):
        async_call(self._connection.execute("SELECT $1;", "1"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        self.assertIs(StatusCode.UNSET, spans[0].status.status_code)

        self.check_span(spans[0])
        self.assertEqual(spans[0].name, "SELECT")
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT $1;"
        )
        self.assertEqual(
            spans[0].attributes["db.statement.parameters"], "('1',)"
        )

    def test_instrumented_fetch_method_with_arguments(self, *_, **__):
        async_call(self._connection.fetch("SELECT $1;", "1"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)

        self.check_span(spans[0])
        self.assertEqual(spans[0].name, "SELECT")
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT $1;"
        )
        self.assertEqual(
            spans[0].attributes["db.statement.parameters"], "('1',)"
        )

    def test_instrumented_executemany_method_with_arguments(self, *_, **__):
        async_call(self._connection.executemany("SELECT $1;", [["1"], ["2"]]))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)

        self.check_span(spans[0])
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT $1;"
        )
        self.assertEqual(
            spans[0].attributes["db.statement.parameters"], "([['1'], ['2']],)"
        )

    def test_instrumented_execute_interface_error_method(self, *_, **__):
        with self.assertRaises(asyncpg.InterfaceError):
            async_call(self._connection.execute("SELECT 42;", 1, 2, 3))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)

        self.check_span(spans[0])
        self.assertEqual(
            spans[0].attributes[SpanAttributes.DB_STATEMENT], "SELECT 42;"
        )
        self.assertEqual(
            spans[0].attributes["db.statement.parameters"], "(1, 2, 3)"
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import time

import mysql.connector
import psycopg2
import pymongo
import pyodbc
import redis

MONGODB_COLLECTION_NAME = "test"
MONGODB_DB_NAME = os.getenv("MONGODB_DB_NAME", "opentelemetry-tests")
MONGODB_HOST = os.getenv("MONGODB_HOST", "localhost")
MONGODB_PORT = int(os.getenv("MONGODB_PORT", "27017"))
MYSQL_DB_NAME = os.getenv("MYSQL_DB_NAME", "opentelemetry-tests")
MYSQL_HOST = os.getenv("MYSQL_HOST", "localhost")
MYSQL_PORT = int(os.getenv("MYSQL_PORT", "3306"))
MYSQL_USER = os.getenv("MYSQL_USER", "testuser")
MYSQL_PASSWORD = os.getenv("MYSQL_PASSWORD", "testpassword")
POSTGRES_DB_NAME = os.getenv("POSTGRESQL_DB_NAME", "opentelemetry-tests")
POSTGRES_HOST = os.getenv("POSTGRESQL_HOST", "localhost")
POSTGRES_PASSWORD = os.getenv("POSTGRESQL_PASSWORD", "testpassword")
POSTGRES_PORT = int(os.getenv("POSTGRESQL_PORT", "5432"))
POSTGRES_USER = os.getenv("POSTGRESQL_USER", "testuser")
REDIS_HOST = os.getenv("REDIS_HOST", "localhost")
REDIS_PORT = int(os.getenv("REDIS_PORT ", "6379"))
MSSQL_DB_NAME = os.getenv("MSSQL_DB_NAME", "opentelemetry-tests")
MSSQL_HOST = os.getenv("MSSQL_HOST", "localhost")
MSSQL_PORT = int(os.getenv("MSSQL_PORT", "1433"))
MSSQL_USER = os.getenv("MSSQL_USER", "sa")
MSSQL_PASSWORD = os.getenv("MSSQL_PASSWORD", "yourStrong(!)Password")
RETRY_COUNT = 8
RETRY_INTERVAL = 5  # Seconds

logger = logging.getLogger(__name__)


def retryable(func):
    def wrapper():
        # Try to connect to DB
        for i in range(RETRY_COUNT):
            try:
                func()
                return
            except Exception as ex:  # pylint: disable=broad-except
                logger.error(
                    "waiting for %s, retry %d/%d [%s]",
                    func.__name__,
                    i + 1,
                    RETRY_COUNT,
                    ex,
                )
            time.sleep(RETRY_INTERVAL)
        raise Exception(f"waiting for {func.__name__} failed")

    return wrapper


@retryable
def check_pymongo_connection():
    client = pymongo.MongoClient(
        MONGODB_HOST, MONGODB_PORT, serverSelectionTimeoutMS=2000
    )
    db = client[MONGODB_DB_NAME]
    collection = db[MONGODB_COLLECTION_NAME]
    collection.find_one()
    client.close()


@retryable
def check_mysql_connection():
    connection = mysql.connector.connect(
        user=MYSQL_USER,
        password=MYSQL_PASSWORD,
        host=MYSQL_HOST,
        port=MYSQL_PORT,
        database=MYSQL_DB_NAME,
    )
    connection.close()


@retryable
def check_postgres_connection():
    connection = psycopg2.connect(
        dbname=POSTGRES_DB_NAME,
        user=POSTGRES_USER,
        password=POSTGRES_PASSWORD,
        host=POSTGRES_HOST,
        port=POSTGRES_PORT,
    )
    connection.close()


@retryable
def check_redis_connection():
    connection = redis.Redis(host=REDIS_HOST, port=REDIS_PORT)
    connection.hgetall("*")


def new_mssql_connection() -> pyodbc.Connection:
    connection = pyodbc.connect(
        f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={MSSQL_HOST},"
        f"{MSSQL_PORT};DATABASE=master;UID={MSSQL_USER};TrustServerCertificate=yes;"
        f"PWD={MSSQL_PASSWORD};",
        autocommit=True,
    )
    return connection


@retryable
def check_mssql_connection():
    conn = new_mssql_connection()
    conn.close()


def setup_mssql_db():
    conn = new_mssql_connection()
    cur = conn.cursor()
    cur.execute(f"CREATE DATABASE [{MSSQL_DB_NAME}]")
    conn.close()


def check_docker_services_availability():
    # Check if Docker services accept connections
    check_pymongo_connection()
    check_mysql_connection()
    check_postgres_connection()
    check_redis_connection()
    check_mssql_connection()
    setup_mssql_db()


check_docker_services_availability()



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
import threading

from sqlalchemy import Column, Integer, String, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import close_all_sessions, scoped_session, sessionmaker

from opentelemetry import trace
from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase

Base = declarative_base()


def _create_engine(engine_args):
    # create a SQLAlchemy engine
    config = dict(engine_args)
    url = config.pop("url")
    return create_engine(url, **config)


class Player(Base):
    """Player entity used to test SQLAlchemy ORM"""

    __tablename__ = "players"

    id = Column(Integer, primary_key=True)
    name = Column(String(20))


class SQLAlchemyTestMixin(TestBase):
    __test__ = False

    """SQLAlchemy test mixin that includes a complete set of tests
    that must be executed for different engine. When a new test (or
    a regression test) should be added to SQLAlchemy test suite, a new
    entry must be appended here so that it will be executed for all
    available and supported engines. If the test is specific to only
    one engine, that test must be added to the specific `TestCase`
    implementation.

    To support a new engine, create a new `TestCase` that inherits from
    `SQLAlchemyTestMixin` and `TestCase`. Then you must define the following
    static class variables:
    * VENDOR: the database vendor name
    * SQL_DB: the `db.type` tag that we expect (it's the name of the database available in the `.env` file)
    * SERVICE: the service that we expect by default
    * ENGINE_ARGS: all arguments required to create the engine

    To check specific tags in each test, you must implement the
    `check_meta(self, span)` method.
    """

    VENDOR = None
    SQL_DB = None
    SERVICE = None
    ENGINE_ARGS = None

    @contextlib.contextmanager
    def connection(self):
        # context manager that provides a connection
        # to the underlying database
        try:
            conn = self.engine.connect()
            yield conn
        finally:
            conn.close()

    def check_meta(self, span):
        """function that can be implemented according to the
        specific engine implementation
        """

    def setUp(self):
        super().setUp()
        # create an engine with the given arguments
        self.engine = _create_engine(self.ENGINE_ARGS)

        # create the database / entities and prepare a session for the test
        Base.metadata.drop_all(bind=self.engine)
        Base.metadata.create_all(self.engine, checkfirst=False)
        self.session = sessionmaker(bind=self.engine)()
        # trace the engine
        SQLAlchemyInstrumentor().instrument(
            engine=self.engine, tracer_provider=self.tracer_provider
        )
        self.memory_exporter.clear()

    def tearDown(self):
        # pylint: disable=invalid-name
        # clear the database and dispose the engine
        self.session.close()
        Base.metadata.drop_all(bind=self.engine)
        self.engine.dispose()
        SQLAlchemyInstrumentor().uninstrument()
        super().tearDown()

    def _check_span(self, span, name):
        if self.SQL_DB:
            name = f"{name} {self.SQL_DB}"
        self.assertEqual(span.name, name)
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_NAME), self.SQL_DB
        )
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)
        self.assertGreater((span.end_time - span.start_time), 0)

    def test_orm_insert(self):
        # ensures that the ORM session is traced
        wayne = Player(id=1, name="wayne")
        self.session.add(wayne)
        self.session.commit()

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one for the query
        self.assertEqual(len(spans), 2)
        span = spans[1]
        stmt = "INSERT INTO players (id, name) VALUES "
        if span.attributes.get(SpanAttributes.DB_SYSTEM) == "sqlite":
            stmt += "(?, ?)"
        else:
            stmt += "(%(id)s, %(name)s)"
        self._check_span(span, "INSERT")
        self.assertIn(
            "INSERT INTO players",
            span.attributes.get(SpanAttributes.DB_STATEMENT),
        )
        self.check_meta(span)

    def test_session_query(self):
        # ensures that the Session queries are traced
        out = list(self.session.query(Player).filter_by(name="wayne"))
        self.assertEqual(len(out), 0)

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one for the query
        self.assertEqual(len(spans), 2)
        span = spans[1]
        stmt = "SELECT players.id AS players_id, players.name AS players_name \nFROM players \nWHERE players.name = "
        if span.attributes.get(SpanAttributes.DB_SYSTEM) == "sqlite":
            stmt += "?"
        else:
            stmt += "%(name_1)s"
        self._check_span(span, "SELECT")
        self.assertIn(
            "SELECT players.id AS players_id, players.name AS players_name \nFROM players \nWHERE players.name",
            span.attributes.get(SpanAttributes.DB_STATEMENT),
        )
        self.check_meta(span)

    def test_engine_connect_execute(self):
        # ensures that engine.connect() is properly traced
        stmt = "SELECT * FROM players"
        with self.connection() as conn:
            rows = conn.execute(stmt).fetchall()
            self.assertEqual(len(rows), 0)

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one for the query
        self.assertEqual(len(spans), 2)
        span = spans[1]
        self._check_span(span, "SELECT")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SELECT * FROM players",
        )
        self.check_meta(span)

    def test_parent(self):
        """Ensure that sqlalchemy works with opentelemetry."""
        stmt = "SELECT * FROM players"
        tracer = self.tracer_provider.get_tracer("sqlalch_svc")

        with tracer.start_as_current_span("sqlalch_op"):
            with self.connection() as conn:
                rows = conn.execute(stmt).fetchall()
                self.assertEqual(len(rows), 0)

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and two for the queries
        self.assertEqual(len(spans), 3)
        _, child_span, parent_span = spans

        # confirm the parenting
        self.assertIsNone(parent_span.parent)
        self.assertIs(child_span.parent, parent_span.get_span_context())

        self.assertEqual(parent_span.name, "sqlalch_op")
        self.assertEqual(parent_span.instrumentation_info.name, "sqlalch_svc")

        self.assertEqual(child_span.name, "SELECT " + self.SQL_DB)

    def test_multithreading(self):
        """Ensure spans are captured correctly in a multithreading scenario

        We also expect no logged warnings about calling end() on an ended span.
        """

        if self.VENDOR == "sqlite":
            return

        def insert_player(session):
            _session = session()
            player = Player(name="Player")
            _session.add(player)
            _session.commit()
            _session.query(Player).all()

        def insert_players(session):
            _session = session()
            players = []
            for player_number in range(3):
                players.append(Player(name=f"Player {player_number}"))
            _session.add_all(players)
            _session.commit()

        session_factory = sessionmaker(bind=self.engine)
        # pylint: disable=invalid-name
        Session = scoped_session(session_factory)
        thread_one = threading.Thread(target=insert_player, args=(Session,))
        thread_two = threading.Thread(target=insert_players, args=(Session,))

        logger = logging.getLogger("opentelemetry.sdk.trace")
        with self.assertRaises(AssertionError):
            with self.assertLogs(logger, level="WARNING"):
                thread_one.start()
                thread_two.start()
                thread_one.join()
                thread_two.join()
                close_all_sessions()

        spans = self.memory_exporter.get_finished_spans()

        # SQLAlchemy 1.4 uses the `execute_values` extension of the psycopg2 dialect to
        # batch inserts together which means `insert_players` only generates one span.
        # See https://docs.sqlalchemy.org/en/14/changelog/migration_14.html#orm-batch-inserts-with-psycopg2-now-batch-statements-with-returning-in-most-cases
        self.assertEqual(
            len(spans), 8 if self.VENDOR not in ["postgresql"] else 6
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import wraps

import pytest

from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.celery import CeleryInstrumentor
from opentelemetry.sdk.trace import TracerProvider, export
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)

REDIS_HOST = os.getenv("REDIS_HOST", "localhost")
REDIS_PORT = int(os.getenv("REDIS_PORT ", "6379"))
REDIS_URL = f"redis://{REDIS_HOST}:{REDIS_PORT}"
BROKER_URL = f"{REDIS_URL}/0"
BACKEND_URL = f"{REDIS_URL}/1"


@pytest.fixture(scope="session")
def celery_config():
    return {"broker_url": BROKER_URL, "result_backend": BACKEND_URL}


@pytest.fixture
def celery_worker_parameters():
    return {
        # See https://github.com/celery/celery/issues/3642#issuecomment-457773294
        "perform_ping_check": False,
    }


@pytest.fixture(autouse=True)
def patch_celery_app(celery_app, celery_worker):
    """Patch task decorator on app fixture to reload worker"""

    # See https://github.com/celery/celery/issues/3642
    def wrap_task(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            result = fn(*args, **kwargs)
            celery_worker.reload()
            return result

        return wrapper

    celery_app.task = wrap_task(celery_app.task)


@pytest.fixture(autouse=True)
def instrument(tracer_provider, memory_exporter):
    CeleryInstrumentor().instrument(tracer_provider=tracer_provider)
    memory_exporter.clear()

    yield

    CeleryInstrumentor().uninstrument()


@pytest.fixture(scope="function")
def tracer_provider(memory_exporter):
    original_tracer_provider = trace_api.get_tracer_provider()

    tracer_provider = TracerProvider()

    span_processor = export.SimpleSpanProcessor(memory_exporter)
    tracer_provider.add_span_processor(span_processor)

    trace_api.set_tracer_provider(tracer_provider)

    yield tracer_provider

    trace_api.set_tracer_provider(original_tracer_provider)


@pytest.fixture(scope="function")
def memory_exporter():
    memory_exporter = InMemorySpanExporter()
    return memory_exporter



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from pymongo import MongoClient

from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.pymongo import PymongoInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase

MONGODB_HOST = os.getenv("MONGODB_HOST", "localhost")
MONGODB_PORT = int(os.getenv("MONGODB_PORT", "27017"))
MONGODB_DB_NAME = os.getenv("MONGODB_DB_NAME", "opentelemetry-tests")
MONGODB_COLLECTION_NAME = "test"


class TestFunctionalPymongo(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        self.instrumentor = PymongoInstrumentor()
        self.instrumentor.instrument()
        self.instrumentor._commandtracer_instance._tracer = self._tracer
        self.instrumentor._commandtracer_instance.capture_statement = True
        client = MongoClient(
            MONGODB_HOST, MONGODB_PORT, serverSelectionTimeoutMS=2000
        )
        db = client[MONGODB_DB_NAME]
        self._collection = db[MONGODB_COLLECTION_NAME]

    def tearDown(self):
        self.instrumentor.uninstrument()
        super().tearDown()

    def validate_spans(self, expected_db_statement):
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        for span in spans:
            if span.name == "rootSpan":
                root_span = span
            else:
                pymongo_span = span
            self.assertIsInstance(span.start_time, int)
            self.assertIsInstance(span.end_time, int)
        self.assertIsNot(root_span, None)
        self.assertIsNot(pymongo_span, None)
        self.assertIsNotNone(pymongo_span.parent)
        self.assertIs(pymongo_span.parent, root_span.get_span_context())
        self.assertIs(pymongo_span.kind, trace_api.SpanKind.CLIENT)
        self.assertEqual(
            pymongo_span.attributes[SpanAttributes.DB_NAME], MONGODB_DB_NAME
        )
        self.assertEqual(
            pymongo_span.attributes[SpanAttributes.NET_PEER_NAME], MONGODB_HOST
        )
        self.assertEqual(
            pymongo_span.attributes[SpanAttributes.NET_PEER_PORT], MONGODB_PORT
        )
        self.assertEqual(
            pymongo_span.attributes[SpanAttributes.DB_MONGODB_COLLECTION],
            MONGODB_COLLECTION_NAME,
        )
        self.assertEqual(
            pymongo_span.attributes[SpanAttributes.DB_STATEMENT],
            expected_db_statement,
        )

    def test_insert(self):
        """Should create a child span for insert"""
        with self._tracer.start_as_current_span("rootSpan"):
            insert_result = self._collection.insert_one(
                {"name": "testName", "value": "testValue"}
            )
            insert_result_id = insert_result.inserted_id

        expected_db_statement = (
            f"insert [{{'name': 'testName', 'value': 'testValue', '_id': "
            f"ObjectId('{insert_result_id}')}}]"
        )
        self.validate_spans(expected_db_statement)

    def test_update(self):
        """Should create a child span for update"""
        with self._tracer.start_as_current_span("rootSpan"):
            self._collection.update_one(
                {"name": "testName"}, {"$set": {"value": "someOtherValue"}}
            )

        expected_db_statement = (
            "update [SON([('q', {'name': 'testName'}), ('u', "
            "{'$set': {'value': 'someOtherValue'}}), ('multi', False), ('upsert', False)])]"
        )
        self.validate_spans(expected_db_statement)

    def test_find(self):
        """Should create a child span for find"""
        with self._tracer.start_as_current_span("rootSpan"):
            self._collection.find_one({"name": "testName"})

        expected_db_statement = "find {'name': 'testName'}"
        self.validate_spans(expected_db_statement)

    def test_delete(self):
        """Should create a child span for delete"""
        with self._tracer.start_as_current_span("rootSpan"):
            self._collection.delete_one({"name": "testName"})

        expected_db_statement = (
            "delete [SON([('q', {'name': 'testName'}), ('limit', 1)])]"
        )
        self.validate_spans(expected_db_statement)

    def test_find_without_capture_statement(self):
        """Should create a child span for find"""
        self.instrumentor._commandtracer_instance.capture_statement = False

        with self._tracer.start_as_current_span("rootSpan"):
            self._collection.find_one({"name": "testName"})

        expected_db_statement = "find"
        self.validate_spans(expected_db_statement)

    def test_uninstrument(self):
        # check that integration is working
        self._collection.find_one()
        spans = self.memory_exporter.get_finished_spans()
        self.memory_exporter.clear()
        self.assertEqual(len(spans), 1)

        # uninstrument and check not new spans are created
        PymongoInstrumentor().uninstrument()
        self._collection.find_one()
        spans = self.memory_exporter.get_finished_spans()
        self.memory_exporter.clear()
        self.assertEqual(len(spans), 0)

        # re-enable and check that it works again
        PymongoInstrumentor().instrument()
        self._collection.find_one()
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sqlalchemy.exc import OperationalError

from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes

from .mixins import SQLAlchemyTestMixin


class SQLiteTestCase(SQLAlchemyTestMixin):
    """TestCase for the SQLite engine"""

    __test__ = True

    VENDOR = "sqlite"
    SQL_DB = ":memory:"
    ENGINE_ARGS = {"url": "sqlite:///:memory:"}

    def test_engine_execute_errors(self):
        # ensures that SQL errors are reported
        stmt = "SELECT * FROM a_wrong_table"
        with pytest.raises(OperationalError):
            with self.connection() as conn:
                conn.execute(stmt).fetchall()

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one span for the query
        self.assertEqual(len(spans), 2)
        span = spans[1]
        # span fields
        self.assertEqual(span.name, "SELECT :memory:")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SELECT * FROM a_wrong_table",
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_NAME), self.SQL_DB
        )
        self.assertTrue((span.end_time - span.start_time) > 0)
        # check the error
        self.assertIs(
            span.status.status_code,
            trace.StatusCode.ERROR,
        )
        self.assertEqual(
            span.status.description, "no such table: a_wrong_table"
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from time import time_ns

import redis
import redis.asyncio
from redis.commands.search.field import (
    TextField,
    VectorField,
)
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.commands.search.query import Query
from redis.exceptions import ResponseError

from opentelemetry import trace
from opentelemetry.instrumentation.redis import RedisInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase


class TestRedisInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.Redis(port=6379)
        self.redis_client.flushall()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

    def tearDown(self):
        RedisInstrumentor().uninstrument()
        super().tearDown()

    def _check_span(self, span, name):
        self.assertEqual(span.name, name)
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_REDIS_DATABASE_INDEX), 0
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_NAME], "localhost"
        )
        self.assertEqual(span.attributes[SpanAttributes.NET_PEER_PORT], 6379)

    def test_long_command_sanitized(self):
        RedisInstrumentor().uninstrument()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

        self.redis_client.mget(*range(2000))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "MGET")
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).startswith(
                "MGET ? ? ? ?"
            )
        )
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).endswith("...")
        )

    def test_long_command(self):
        self.redis_client.mget(*range(1000))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "MGET")
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).startswith(
                "MGET ? ? ? ?"
            )
        )
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).endswith("...")
        )

    def test_basics_sanitized(self):
        RedisInstrumentor().uninstrument()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

        self.assertIsNone(self.redis_client.get("cheese"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )
        self.assertEqual(span.attributes.get("db.redis.args_length"), 2)

    def test_basics(self):
        self.assertIsNone(self.redis_client.get("cheese"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )
        self.assertEqual(span.attributes.get("db.redis.args_length"), 2)

    def test_pipeline_traced_sanitized(self):
        RedisInstrumentor().uninstrument()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

        with self.redis_client.pipeline(transaction=False) as pipeline:
            pipeline.set("blah", 32)
            pipeline.rpush("foo", "éé")
            pipeline.hgetall("xxx")
            pipeline.execute()

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "SET RPUSH HGETALL")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SET ? ?\nRPUSH ? ?\nHGETALL ?",
        )
        self.assertEqual(span.attributes.get("db.redis.pipeline_length"), 3)

    def test_pipeline_traced(self):
        with self.redis_client.pipeline(transaction=False) as pipeline:
            pipeline.set("blah", 32)
            pipeline.rpush("foo", "éé")
            pipeline.hgetall("xxx")
            pipeline.execute()

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "SET RPUSH HGETALL")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SET ? ?\nRPUSH ? ?\nHGETALL ?",
        )
        self.assertEqual(span.attributes.get("db.redis.pipeline_length"), 3)

    def test_pipeline_immediate_sanitized(self):
        RedisInstrumentor().uninstrument()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

        with self.redis_client.pipeline() as pipeline:
            pipeline.set("a", 1)
            pipeline.immediate_execute_command("SET", "b", 2)
            pipeline.execute()

        spans = self.memory_exporter.get_finished_spans()
        # expecting two separate spans here, rather than a
        # single span for the whole pipeline
        self.assertEqual(len(spans), 2)
        span = spans[0]
        self._check_span(span, "SET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "SET ? ?"
        )

    def test_pipeline_immediate(self):
        with self.redis_client.pipeline() as pipeline:
            pipeline.set("a", 1)
            pipeline.immediate_execute_command("SET", "b", 2)
            pipeline.execute()

        spans = self.memory_exporter.get_finished_spans()
        # expecting two separate spans here, rather than a
        # single span for the whole pipeline
        self.assertEqual(len(spans), 2)
        span = spans[0]
        self._check_span(span, "SET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "SET ? ?"
        )

    def test_parent(self):
        """Ensure OpenTelemetry works with redis."""
        ot_tracer = trace.get_tracer("redis_svc")

        with ot_tracer.start_as_current_span("redis_get"):
            self.assertIsNone(self.redis_client.get("cheese"))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        child_span, parent_span = spans[0], spans[1]

        # confirm the parenting
        self.assertIsNone(parent_span.parent)
        self.assertIs(child_span.parent, parent_span.get_span_context())

        self.assertEqual(parent_span.name, "redis_get")
        self.assertEqual(parent_span.instrumentation_info.name, "redis_svc")

        self.assertEqual(child_span.name, "GET")


class TestRedisClusterInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.cluster.RedisCluster(
            host="localhost", port=7000
        )
        self.redis_client.flushall()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

    def tearDown(self):
        super().tearDown()
        RedisInstrumentor().uninstrument()

    def _check_span(self, span, name):
        self.assertEqual(span.name, name)
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)

    def test_basics(self):
        self.assertIsNone(self.redis_client.get("cheese"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )
        self.assertEqual(span.attributes.get("db.redis.args_length"), 2)

    def test_pipeline_traced(self):
        with self.redis_client.pipeline(transaction=False) as pipeline:
            pipeline.set("blah", 32)
            pipeline.rpush("foo", "éé")
            pipeline.hgetall("xxx")
            pipeline.execute()

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "SET RPUSH HGETALL")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SET ? ?\nRPUSH ? ?\nHGETALL ?",
        )
        self.assertEqual(span.attributes.get("db.redis.pipeline_length"), 3)

    def test_parent(self):
        """Ensure OpenTelemetry works with redis."""
        ot_tracer = trace.get_tracer("redis_svc")

        with ot_tracer.start_as_current_span("redis_get"):
            self.assertIsNone(self.redis_client.get("cheese"))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        child_span, parent_span = spans[0], spans[1]

        # confirm the parenting
        self.assertIsNone(parent_span.parent)
        self.assertIs(child_span.parent, parent_span.get_span_context())

        self.assertEqual(parent_span.name, "redis_get")
        self.assertEqual(parent_span.instrumentation_info.name, "redis_svc")

        self.assertEqual(child_span.name, "GET")


def async_call(coro):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)


class TestAsyncRedisInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.asyncio.Redis(port=6379)
        async_call(self.redis_client.flushall())
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

    def tearDown(self):
        RedisInstrumentor().uninstrument()
        super().tearDown()

    def _check_span(self, span, name):
        self.assertEqual(span.name, name)
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_REDIS_DATABASE_INDEX), 0
        )
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_NAME], "localhost"
        )
        self.assertEqual(span.attributes[SpanAttributes.NET_PEER_PORT], 6379)

    def test_long_command(self):
        async_call(self.redis_client.mget(*range(1000)))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "MGET")
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).startswith(
                "MGET ? ? ? ?"
            )
        )
        self.assertTrue(
            span.attributes.get(SpanAttributes.DB_STATEMENT).endswith("...")
        )

    def test_basics(self):
        self.assertIsNone(async_call(self.redis_client.get("cheese")))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )
        self.assertEqual(span.attributes.get("db.redis.args_length"), 2)

    def test_execute_command_traced_full_time(self):
        """Command should be traced for coroutine execution time, not creation time."""
        coro_created_time = None
        finish_time = None

        async def pipeline_simple():
            nonlocal coro_created_time
            nonlocal finish_time

            # delay coroutine creation from coroutine execution
            coro = self.redis_client.get("foo")
            coro_created_time = time_ns()
            await coro
            finish_time = time_ns()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertTrue(span.start_time > coro_created_time)
        self.assertTrue(span.end_time < finish_time)

    def test_pipeline_traced(self):
        async def pipeline_simple():
            async with self.redis_client.pipeline(
                transaction=False
            ) as pipeline:
                pipeline.set("blah", 32)
                pipeline.rpush("foo", "éé")
                pipeline.hgetall("xxx")
                await pipeline.execute()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "SET RPUSH HGETALL")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SET ? ?\nRPUSH ? ?\nHGETALL ?",
        )
        self.assertEqual(span.attributes.get("db.redis.pipeline_length"), 3)

    def test_pipeline_traced_full_time(self):
        """Command should be traced for coroutine execution time, not creation time."""
        coro_created_time = None
        finish_time = None

        async def pipeline_simple():
            async with self.redis_client.pipeline(
                transaction=False
            ) as pipeline:
                nonlocal coro_created_time
                nonlocal finish_time
                pipeline.set("blah", 32)
                pipeline.rpush("foo", "éé")
                pipeline.hgetall("xxx")

                # delay coroutine creation from coroutine execution
                coro = pipeline.execute()
                coro_created_time = time_ns()
                await coro
                finish_time = time_ns()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertTrue(span.start_time > coro_created_time)
        self.assertTrue(span.end_time < finish_time)

    def test_pipeline_immediate(self):
        async def pipeline_immediate():
            async with self.redis_client.pipeline() as pipeline:
                pipeline.set("a", 1)
                await pipeline.immediate_execute_command("SET", "b", 2)
                await pipeline.execute()

        async_call(pipeline_immediate())

        spans = self.memory_exporter.get_finished_spans()
        # expecting two separate spans here, rather than a
        # single span for the whole pipeline
        self.assertEqual(len(spans), 2)
        span = spans[0]
        self._check_span(span, "SET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "SET ? ?"
        )

    def test_pipeline_immediate_traced_full_time(self):
        """Command should be traced for coroutine execution time, not creation time."""
        coro_created_time = None
        finish_time = None

        async def pipeline_simple():
            async with self.redis_client.pipeline(
                transaction=False
            ) as pipeline:
                nonlocal coro_created_time
                nonlocal finish_time
                pipeline.set("a", 1)

                # delay coroutine creation from coroutine execution
                coro = pipeline.immediate_execute_command("SET", "b", 2)
                coro_created_time = time_ns()
                await coro
                finish_time = time_ns()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertTrue(span.start_time > coro_created_time)
        self.assertTrue(span.end_time < finish_time)

    def test_parent(self):
        """Ensure OpenTelemetry works with redis."""
        ot_tracer = trace.get_tracer("redis_svc")

        with ot_tracer.start_as_current_span("redis_get"):
            self.assertIsNone(async_call(self.redis_client.get("cheese")))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        child_span, parent_span = spans[0], spans[1]

        # confirm the parenting
        self.assertIsNone(parent_span.parent)
        self.assertIs(child_span.parent, parent_span.get_span_context())

        self.assertEqual(parent_span.name, "redis_get")
        self.assertEqual(parent_span.instrumentation_info.name, "redis_svc")

        self.assertEqual(child_span.name, "GET")


class TestAsyncRedisClusterInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.asyncio.cluster.RedisCluster(
            host="localhost", port=7000
        )
        async_call(self.redis_client.flushall())
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

    def tearDown(self):
        super().tearDown()
        RedisInstrumentor().uninstrument()

    def _check_span(self, span, name):
        self.assertEqual(span.name, name)
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)

    def test_basics(self):
        self.assertIsNone(async_call(self.redis_client.get("cheese")))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )
        self.assertEqual(span.attributes.get("db.redis.args_length"), 2)

    def test_execute_command_traced_full_time(self):
        """Command should be traced for coroutine execution time, not creation time."""
        coro_created_time = None
        finish_time = None

        async def pipeline_simple():
            nonlocal coro_created_time
            nonlocal finish_time

            # delay coroutine creation from coroutine execution
            coro = self.redis_client.get("foo")
            coro_created_time = time_ns()
            await coro
            finish_time = time_ns()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertTrue(span.start_time > coro_created_time)
        self.assertTrue(span.end_time < finish_time)

    def test_pipeline_traced(self):
        async def pipeline_simple():
            async with self.redis_client.pipeline(
                transaction=False
            ) as pipeline:
                pipeline.set("blah", 32)
                pipeline.rpush("foo", "éé")
                pipeline.hgetall("xxx")
                await pipeline.execute()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "SET RPUSH HGETALL")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SET ? ?\nRPUSH ? ?\nHGETALL ?",
        )
        self.assertEqual(span.attributes.get("db.redis.pipeline_length"), 3)

    def test_pipeline_traced_full_time(self):
        """Command should be traced for coroutine execution time, not creation time."""
        coro_created_time = None
        finish_time = None

        async def pipeline_simple():
            async with self.redis_client.pipeline(
                transaction=False
            ) as pipeline:
                nonlocal coro_created_time
                nonlocal finish_time
                pipeline.set("blah", 32)
                pipeline.rpush("foo", "éé")
                pipeline.hgetall("xxx")

                # delay coroutine creation from coroutine execution
                coro = pipeline.execute()
                coro_created_time = time_ns()
                await coro
                finish_time = time_ns()

        async_call(pipeline_simple())

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertTrue(span.start_time > coro_created_time)
        self.assertTrue(span.end_time < finish_time)

    def test_parent(self):
        """Ensure OpenTelemetry works with redis."""
        ot_tracer = trace.get_tracer("redis_svc")

        with ot_tracer.start_as_current_span("redis_get"):
            self.assertIsNone(async_call(self.redis_client.get("cheese")))

        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        child_span, parent_span = spans[0], spans[1]

        # confirm the parenting
        self.assertIsNone(parent_span.parent)
        self.assertIs(child_span.parent, parent_span.get_span_context())

        self.assertEqual(parent_span.name, "redis_get")
        self.assertEqual(parent_span.instrumentation_info.name, "redis_svc")

        self.assertEqual(child_span.name, "GET")


class TestRedisDBIndexInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.Redis(port=6379, db=10)
        self.redis_client.flushall()
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)

    def tearDown(self):
        RedisInstrumentor().uninstrument()
        super().tearDown()

    def _check_span(self, span, name):
        self.assertEqual(span.name, name)
        self.assertIs(span.status.status_code, trace.StatusCode.UNSET)
        self.assertEqual(
            span.attributes[SpanAttributes.NET_PEER_NAME], "localhost"
        )
        self.assertEqual(span.attributes[SpanAttributes.NET_PEER_PORT], 6379)
        self.assertEqual(
            span.attributes[SpanAttributes.DB_REDIS_DATABASE_INDEX], 10
        )

    def test_get(self):
        self.assertIsNone(self.redis_client.get("foo"))
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 1)
        span = spans[0]
        self._check_span(span, "GET")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT), "GET ?"
        )


class TestRedisearchInstrument(TestBase):
    def setUp(self):
        super().setUp()
        self.redis_client = redis.Redis(port=6379)
        self.redis_client.flushall()
        self.embedding_dim = 256
        RedisInstrumentor().instrument(tracer_provider=self.tracer_provider)
        self.prepare_data()
        self.create_index()

    def tearDown(self):
        RedisInstrumentor().uninstrument()
        super().tearDown()

    def prepare_data(self):
        try:
            self.redis_client.ft("idx:test_vss").dropindex(True)
        except ResponseError:
            print("No such index")
        item = {
            "name": "test",
            "value": "test_value",
            "embeddings": [0.1] * 256,
        }
        pipeline = self.redis_client.pipeline()
        pipeline.json().set("test:001", "$", item)
        res = pipeline.execute()
        assert False not in res

    def create_index(self):
        schema = (
            TextField("$.name", no_stem=True, as_name="name"),
            TextField("$.value", no_stem=True, as_name="value"),
            VectorField(
                "$.embeddings",
                "FLAT",
                {
                    "TYPE": "FLOAT32",
                    "DIM": self.embedding_dim,
                    "DISTANCE_METRIC": "COSINE",
                },
                as_name="vector",
            ),
        )
        definition = IndexDefinition(
            prefix=["test:"], index_type=IndexType.JSON
        )
        res = self.redis_client.ft("idx:test_vss").create_index(
            fields=schema, definition=definition
        )
        assert "OK" in str(res)

    def test_redis_create_index(self):
        spans = self.memory_exporter.get_finished_spans()
        span = next(
            span for span in spans if span.name == "redis.create_index"
        )
        assert "redis.create_index.fields" in span.attributes

    def test_redis_query(self):
        query = "@name:test"
        self.redis_client.ft("idx:test_vss").search(Query(query))

        spans = self.memory_exporter.get_finished_spans()
        span = next(span for span in spans if span.name == "redis.search")

        assert span.attributes.get("redis.search.query") == query
        assert span.attributes.get("redis.search.total") == 1



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest
from sqlalchemy.exc import ProgrammingError

from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes

from .mixins import SQLAlchemyTestMixin

MYSQL_CONFIG = {
    "host": "127.0.0.1",
    "port": int(os.getenv("TEST_MYSQL_PORT", "3306")),
    "user": os.getenv("TEST_MYSQL_USER", "testuser"),
    "password": os.getenv("TEST_MYSQL_PASSWORD", "testpassword"),
    "database": os.getenv("TEST_MYSQL_DATABASE", "opentelemetry-tests"),
}


class MysqlConnectorTestCase(SQLAlchemyTestMixin):
    """TestCase for mysql-connector engine"""

    __test__ = True

    VENDOR = "mysql"
    SQL_DB = "opentelemetry-tests"
    ENGINE_ARGS = {
        "url": "mysql+mysqlconnector://%(user)s:%(password)s@%(host)s:%(port)s/%(database)s"
        % MYSQL_CONFIG
    }

    def check_meta(self, span):
        # check database connection tags
        self.assertEqual(
            span.attributes.get(SpanAttributes.NET_PEER_NAME),
            MYSQL_CONFIG["host"],
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.NET_PEER_PORT),
            MYSQL_CONFIG["port"],
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_NAME),
            MYSQL_CONFIG["database"],
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_USER), MYSQL_CONFIG["user"]
        )

    def test_engine_execute_errors(self):
        # ensures that SQL errors are reported
        with pytest.raises(ProgrammingError):
            with self.connection() as conn:
                conn.execute("SELECT * FROM a_wrong_table").fetchall()

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one for the query
        self.assertEqual(len(spans), 2)
        self.check_meta(spans[0])
        span = spans[1]
        # span fields
        self.assertEqual(span.name, "SELECT opentelemetry-tests")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SELECT * FROM a_wrong_table",
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_NAME), self.SQL_DB
        )
        self.check_meta(span)
        self.assertTrue(span.end_time - span.start_time > 0)
        # check the error
        self.assertIs(
            span.status.status_code,
            trace.StatusCode.ERROR,
        )
        self.assertIn("a_wrong_table", span.status.description)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import psycopg2
import pytest
from sqlalchemy.exc import ProgrammingError

from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes

from .mixins import SQLAlchemyTestMixin

POSTGRES_CONFIG = {
    "host": "127.0.0.1",
    "port": int(os.getenv("TEST_POSTGRES_PORT", "5432")),
    "user": os.getenv("TEST_POSTGRES_USER", "testuser"),
    "password": os.getenv("TEST_POSTGRES_PASSWORD", "testpassword"),
    "dbname": os.getenv("TEST_POSTGRES_DB", "opentelemetry-tests"),
}


class PostgresTestCase(SQLAlchemyTestMixin):
    """TestCase for Postgres Engine"""

    __test__ = True

    VENDOR = "postgresql"
    SQL_DB = "opentelemetry-tests"
    ENGINE_ARGS = {
        "url": "postgresql://%(user)s:%(password)s@%(host)s:%(port)s/%(dbname)s"
        % POSTGRES_CONFIG
    }

    def check_meta(self, span):
        # check database connection tags
        self.assertEqual(
            span.attributes.get(SpanAttributes.NET_PEER_NAME),
            POSTGRES_CONFIG["host"],
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.NET_PEER_PORT),
            POSTGRES_CONFIG["port"],
        )

    def test_engine_execute_errors(self):
        # ensures that SQL errors are reported
        with pytest.raises(ProgrammingError):
            with self.connection() as conn:
                conn.execute("SELECT * FROM a_wrong_table").fetchall()

        spans = self.memory_exporter.get_finished_spans()
        # one span for the connection and one for the query
        self.assertEqual(len(spans), 2)
        span = spans[1]
        # span fields
        self.assertEqual(span.name, "SELECT opentelemetry-tests")
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_STATEMENT),
            "SELECT * FROM a_wrong_table",
        )
        self.assertEqual(
            span.attributes.get(SpanAttributes.DB_NAME), self.SQL_DB
        )
        self.check_meta(span)
        self.assertTrue(span.end_time - span.start_time > 0)
        # check the error
        self.assertIs(
            span.status.status_code,
            trace.StatusCode.ERROR,
        )
        self.assertIn("a_wrong_table", span.status.description)


class PostgresCreatorTestCase(PostgresTestCase):
    """TestCase for Postgres Engine that includes the same tests set
    of `PostgresTestCase`, but it uses a specific `creator` function.
    """

    VENDOR = "postgresql"
    SQL_DB = "opentelemetry-tests"
    ENGINE_ARGS = {
        "url": "postgresql://",
        "creator": lambda: psycopg2.connect(**POSTGRES_CONFIG),
    }


class PostgresMetricsTestCase(PostgresTestCase):
    __test__ = True

    VENDOR = "postgresql"
    SQL_DB = "opentelemetry-tests"
    ENGINE_ARGS = {
        "url": "postgresql://%(user)s:%(password)s@%(host)s:%(port)s/%(dbname)s"
        % POSTGRES_CONFIG
    }

    def test_metrics_pool_name(self):
        with self.connection() as conn:
            conn.execute("SELECT 1 + 1").fetchall()

        pool_name = "{}://{}:{}/{}".format(
            self.VENDOR,
            POSTGRES_CONFIG["host"],
            POSTGRES_CONFIG["port"],
            self.SQL_DB,
        )
        metrics = self.get_sorted_metrics()
        self.assertEqual(len(metrics), 1)
        self.assert_metric_expected(
            metrics[0],
            [
                self.create_number_data_point(
                    value=0,
                    attributes={"pool.name": pool_name, "state": "idle"},
                ),
                self.create_number_data_point(
                    value=0,
                    attributes={"pool.name": pool_name, "state": "used"},
                ),
            ],
        )



# Copyright 2020, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
from test_psycopg_functional import (
    POSTGRES_DB_NAME,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USER,
)

from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor
from opentelemetry.test.test_base import TestBase


class TestFunctionalPsycopg(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        Psycopg2Instrumentor().instrument(enable_commenter=True)
        self._connection = psycopg2.connect(
            dbname=POSTGRES_DB_NAME,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
            host=POSTGRES_HOST,
            port=POSTGRES_PORT,
        )
        self._connection.set_session(autocommit=True)
        self._cursor = self._connection.cursor()

    def tearDown(self):
        self._cursor.close()
        self._connection.close()
        Psycopg2Instrumentor().uninstrument()
        super().tearDown()

    def test_commenter_enabled(self):
        self._cursor.execute("SELECT  1;")
        self.assertRegex(
            self._cursor.query.decode("ascii"),
            r"SELECT  1 /\*db_driver='psycopg2(.*)',dbapi_level='\d.\d',dbapi_threadsafety=\d,driver_paramstyle=(.*),libpq_version=\d*,traceparent='\d{1,2}-[a-zA-Z0-9_]{32}-[a-zA-Z0-9_]{16}-\d{1,2}'\*/;",
        )



# Copyright 2020, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import os

import aiopg
import psycopg2
import pytest

from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.aiopg import AiopgInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase

POSTGRES_HOST = os.getenv("POSTGRESQL_HOST", "127.0.0.1")
POSTGRES_PORT = int(os.getenv("POSTGRESQL_PORT", "5432"))
POSTGRES_DB_NAME = os.getenv("POSTGRESQL_DB_NAME", "opentelemetry-tests")
POSTGRES_PASSWORD = os.getenv("POSTGRESQL_PASSWORD", "testpassword")
POSTGRES_USER = os.getenv("POSTGRESQL_USER", "testuser")


def async_call(coro):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)


class TestFunctionalAiopgConnect(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        AiopgInstrumentor().instrument(tracer_provider=self.tracer_provider)
        self._connection = async_call(
            aiopg.connect(
                dbname=POSTGRES_DB_NAME,
                user=POSTGRES_USER,
                password=POSTGRES_PASSWORD,
                host=POSTGRES_HOST,
                port=POSTGRES_PORT,
            )
        )
        self._cursor = async_call(self._connection.cursor())

    def tearDown(self):
        self._cursor.close()
        self._connection.close()
        AiopgInstrumentor().uninstrument()
        super().tearDown()

    def validate_spans(self, span_name):
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        for span in spans:
            if span.name == "rootSpan":
                root_span = span
            else:
                child_span = span
            self.assertIsInstance(span.start_time, int)
            self.assertIsInstance(span.end_time, int)
        self.assertIsNotNone(root_span)
        self.assertIsNotNone(child_span)
        self.assertEqual(root_span.name, "rootSpan")
        self.assertEqual(child_span.name, span_name)
        self.assertIsNotNone(child_span.parent)
        self.assertIs(child_span.parent, root_span.get_span_context())
        self.assertIs(child_span.kind, trace_api.SpanKind.CLIENT)
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_SYSTEM], "postgresql"
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_NAME], POSTGRES_DB_NAME
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_USER], POSTGRES_USER
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.NET_PEER_NAME], POSTGRES_HOST
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.NET_PEER_PORT], POSTGRES_PORT
        )

    def test_execute(self):
        """Should create a child span for execute method"""
        stmt = "CREATE TABLE IF NOT EXISTS test (id integer)"
        with self._tracer.start_as_current_span("rootSpan"):
            async_call(self._cursor.execute(stmt))
        self.validate_spans("CREATE")

    def test_executemany(self):
        """Should create a child span for executemany"""
        stmt = "INSERT INTO test (id) VALUES (%s)"
        with pytest.raises(psycopg2.ProgrammingError):
            with self._tracer.start_as_current_span("rootSpan"):
                data = (("1",), ("2",), ("3",))
                async_call(self._cursor.executemany(stmt, data))
            self.validate_spans("INSERT")

    def test_callproc(self):
        """Should create a child span for callproc"""
        with self._tracer.start_as_current_span("rootSpan"), self.assertRaises(
            Exception
        ):
            async_call(self._cursor.callproc("test", ()))
            self.validate_spans("test")


class TestFunctionalAiopgCreatePool(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        AiopgInstrumentor().instrument(tracer_provider=self.tracer_provider)
        self._dsn = (
            f"dbname='{POSTGRES_DB_NAME}' user='{POSTGRES_USER}' password='{POSTGRES_PASSWORD}'"
            f" host='{POSTGRES_HOST}' port='{POSTGRES_PORT}'"
        )
        self._pool = async_call(
            aiopg.create_pool(
                dbname=POSTGRES_DB_NAME,
                user=POSTGRES_USER,
                password=POSTGRES_PASSWORD,
                host=POSTGRES_HOST,
                port=POSTGRES_PORT,
            )
        )
        self._connection = async_call(self._pool.acquire())
        self._cursor = async_call(self._connection.cursor())

    def tearDown(self):
        self._cursor.close()
        self._connection.close()
        self._pool.close()
        AiopgInstrumentor().uninstrument()
        super().tearDown()

    def validate_spans(self, span_name):
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        for span in spans:
            if span.name == "rootSpan":
                root_span = span
            else:
                child_span = span
            self.assertIsInstance(span.start_time, int)
            self.assertIsInstance(span.end_time, int)
        self.assertIsNotNone(root_span)
        self.assertIsNotNone(child_span)
        self.assertEqual(root_span.name, "rootSpan")
        self.assertEqual(child_span.name, span_name)
        self.assertIsNotNone(child_span.parent)
        self.assertIs(child_span.parent, root_span.get_span_context())
        self.assertIs(child_span.kind, trace_api.SpanKind.CLIENT)
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_SYSTEM], "postgresql"
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_NAME], POSTGRES_DB_NAME
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.DB_USER], POSTGRES_USER
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.NET_PEER_NAME], POSTGRES_HOST
        )
        self.assertEqual(
            child_span.attributes[SpanAttributes.NET_PEER_PORT], POSTGRES_PORT
        )

    def test_execute(self):
        """Should create a child span for execute method"""
        stmt = "CREATE TABLE IF NOT EXISTS test (id integer)"
        with self._tracer.start_as_current_span("rootSpan"):
            async_call(self._cursor.execute(stmt))
        self.validate_spans("CREATE")

    def test_executemany(self):
        """Should create a child span for executemany"""
        stmt = "INSERT INTO test (id) VALUES (%s)"
        with pytest.raises(psycopg2.ProgrammingError):
            with self._tracer.start_as_current_span("rootSpan"):
                data = (("1",), ("2",), ("3",))
                async_call(self._cursor.executemany(stmt, data))
            self.validate_spans("INSERT")

    def test_callproc(self):
        """Should create a child span for callproc"""
        with self._tracer.start_as_current_span("rootSpan"), self.assertRaises(
            Exception
        ):
            async_call(self._cursor.callproc("test", ()))
            self.validate_spans("test")

    def test_instrumented_pool_with_multiple_acquires(self, *_, **__):
        async def double_acquire():
            pool = await aiopg.create_pool(dsn=self._dsn)
            async with pool.acquire() as conn:
                async with conn.cursor() as cursor:
                    query = "SELECT 1"
                    await cursor.execute(query)
            async with pool.acquire() as conn:
                async with conn.cursor() as cursor:
                    query = "SELECT 1"
                    await cursor.execute(query)

        async_call(double_acquire())
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import mysql.connector

from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.mysql import MySQLInstrumentor
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.test.test_base import TestBase

MYSQL_USER = os.getenv("MYSQL_USER", "testuser")
MYSQL_PASSWORD = os.getenv("MYSQL_PASSWORD", "testpassword")
MYSQL_HOST = os.getenv("MYSQL_HOST", "localhost")
MYSQL_PORT = int(os.getenv("MYSQL_PORT", "3306"))
MYSQL_DB_NAME = os.getenv("MYSQL_DB_NAME", "opentelemetry-tests")


class TestFunctionalMysql(TestBase):
    def setUp(self):
        super().setUp()
        self._tracer = self.tracer_provider.get_tracer(__name__)
        MySQLInstrumentor().instrument()
        self._connection = mysql.connector.connect(
            user=MYSQL_USER,
            password=MYSQL_PASSWORD,
            host=MYSQL_HOST,
            port=MYSQL_PORT,
            database=MYSQL_DB_NAME,
        )
        self._cursor = self._connection.cursor()

    def tearDown(self):
        self._cursor.close()
        self._connection.close()
        MySQLInstrumentor().uninstrument()
        super().tearDown()

    def validate_spans(self, span_name):
        spans = self.memory_exporter.get_finished_spans()
        self.assertEqual(len(spans), 2)
        for span in spans:
            if span.name == "rootSpan":
                root_span = span
            else:
                db_span = span
            self.assertIsInstance(span.start_time, int)
            self.assertIsInstance(span.end_time, int)
        self.assertIsNotNone(root_span)
        self.assertIsNotNone(db_span)
        self.assertEqual(root_span.name, "rootSpan")
        self.assertEqual(db_span.name, span_name)
        self.assertIsNotNone(db_span.parent)
        self.assertIs(db_span.parent, root_span.get_span_context())
        self.assertIs(db_span.kind, trace_api.SpanKind.CLIENT)
        self.assertEqual(db_span.attributes[SpanAttributes.DB_SYSTEM], "mysql")
        self.assertEqual(
            db_span.attributes[SpanAttributes.DB_NAME], MYSQL_DB_NAME
        )
        self.assertEqual(
            db_span.attributes[SpanAttributes.DB_USER], MYSQL_USER
        )
        self.assertEqual(
            db_span.attributes[SpanAttributes.NET_PEER_NAME], MYSQL_HOST
        )
        self.assertEqual(
            db_span.attributes[SpanAttributes.NET_PEER_PORT], MYSQL_PORT
        )

    def test_execute(self):
        """Should create a child span for execute"""
        stmt = "CREATE TABLE IF NOT EXISTS test (id INT)"
        with self._tracer.start_as_current_span("rootSpan"):
            self._cursor.execute(stmt)
        self.validate_spans("CREATE")

    def test_execute_with_connection_context_manager(self):
        """Should create a child span for execute with connection context"""
        stmt = "CREATE TABLE IF NOT EXISTS test (id INT)"
        with self._tracer.start_as_current_span("rootSpan"):
            with self._connection as conn:
                cursor = conn.cursor()
                cursor.execute(stmt)
        self.validate_spans("CREATE")

    def test_execute_with_cursor_context_manager(self):
        """Should create a child span for execute with cursor context"""
        stmt = "CREATE TABLE IF NOT EXISTS test (id INT)"
        with self._tracer.start_as_current_span("rootSpan"):
            with self._connection.cursor() as cursor:
                cursor.execute(stmt)
        self.validate_spans("CREATE")

    def test_executemany(self):
        """Should create a child span for executemany"""
        stmt = "INSERT INTO test (id) VALUES (%s)"
        with self._tracer.start_as_current_span("rootSpan"):
            data = (("1",), ("2",), ("3",))
            self._cursor.executemany(stmt, data)
        self.validate_spans("INSERT")

    def test_callproc(self):
        """Should create a child span for callproc"""
        with self._tracer.start_as_current_span("rootSpan"), self.assertRaises(
            Exception
        ):
            self._cursor.callproc("test", ())
            self.validate_spans("test")



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import celery
from celery.exceptions import Retry
from pytest import mark

import opentelemetry.instrumentation.celery
from opentelemetry import trace as trace_api
from opentelemetry.instrumentation.celery import CeleryInstrumentor
from opentelemetry.sdk import resources
from opentelemetry.sdk.trace import TracerProvider, export
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import StatusCode

# set a high timeout for async executions due to issues in CI
ASYNC_GET_TIMEOUT = 120


class MyException(Exception):
    pass


@mark.skip(reason="inconsistent test results")
def test_instrumentation_info(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    result = fn_task.apply_async()
    assert result.get(timeout=ASYNC_GET_TIMEOUT) == 42

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 2

    async_span, run_span = spans

    assert run_span.parent == async_span.context
    assert run_span.parent.span_id == async_span.context.span_id
    assert run_span.context.trace_id == async_span.context.trace_id

    assert async_span.instrumentation_info.name == "apply_async/{}".format(
        opentelemetry.instrumentation.celery.__name__
    )
    assert async_span.instrumentation_info.version == "apply_async/{}".format(
        opentelemetry.instrumentation.celery.__version__
    )
    assert run_span.instrumentation_info.name == "run/{}".format(
        opentelemetry.instrumentation.celery.__name__
    )
    assert run_span.instrumentation_info.version == "run/{}".format(
        opentelemetry.instrumentation.celery.__version__
    )


def test_fn_task_run(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    t = fn_task.run()
    assert t == 42

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 0


def test_fn_task_call(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    t = fn_task()
    assert t == 42

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 0


def test_fn_task_apply(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    t = fn_task.apply()
    assert t.successful() is True
    assert t.result == 42

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is True
    assert span.name == "run/test_celery_functional.fn_task"
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID) == t.task_id
    )
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task"
    )
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "SUCCESS"


def test_fn_task_apply_bind(celery_app, memory_exporter):
    @celery_app.task(bind=True)
    def fn_task(self):
        return self

    t = fn_task.apply()
    assert t.successful() is True
    assert "fn_task" in t.result.name

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is True
    assert span.name == "run/test_celery_functional.fn_task"
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID) == t.task_id
    )
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task"
    )
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "SUCCESS"


@mark.skip(reason="inconsistent test results")
def test_fn_task_apply_async(celery_app, memory_exporter):
    @celery_app.task
    def fn_task_parameters(user, force_logout=False):
        return (user, force_logout)

    result = fn_task_parameters.apply_async(
        args=["user"], kwargs={"force_logout": True}
    )
    assert result.get(timeout=ASYNC_GET_TIMEOUT) == ["user", True]

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 2

    async_span, run_span = spans

    assert run_span.context.trace_id != async_span.context.trace_id

    assert async_span.status.is_ok is True
    assert (
        async_span.name
        == "apply_async/test_celery_functional.fn_task_parameters"
    )
    assert async_span.attributes.get("celery.action") == "apply_async"
    assert (
        async_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert (
        async_span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task_parameters"
    )

    assert run_span.status.is_ok is True
    assert run_span.name == "test_celery_functional.fn_task_parameters"
    assert run_span.attributes.get("celery.action") == "run"
    assert run_span.attributes.get("celery.state") == "SUCCESS"
    assert (
        run_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert (
        run_span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task_parameters"
    )


@mark.skip(reason="inconsistent test results")
def test_concurrent_delays(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    results = [fn_task.delay() for _ in range(100)]

    for result in results:
        assert result.get(timeout=ASYNC_GET_TIMEOUT) == 42

    spans = memory_exporter.get_finished_spans()

    assert len(spans) == 200


@mark.skip(reason="inconsistent test results")
def test_fn_task_delay(celery_app, memory_exporter):
    @celery_app.task
    def fn_task_parameters(user, force_logout=False):
        return (user, force_logout)

    result = fn_task_parameters.delay("user", force_logout=True)
    assert result.get(timeout=ASYNC_GET_TIMEOUT) == ["user", True]

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 2

    async_span, run_span = spans

    assert run_span.context.trace_id != async_span.context.trace_id

    assert async_span.status.is_ok is True
    assert (
        async_span.name
        == "apply_async/test_celery_functional.fn_task_parameters"
    )
    assert async_span.attributes.get("celery.action") == "apply_async"
    assert (
        async_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert (
        async_span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task_parameters"
    )

    assert run_span.status.is_ok is True
    assert run_span.name == "run/test_celery_functional.fn_task_parameters"
    assert run_span.attributes.get("celery.action") == "run"
    assert run_span.attributes.get("celery.state") == "SUCCESS"
    assert (
        run_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert (
        run_span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_task_parameters"
    )
    assert len(run_span.events) == 0


def test_fn_exception(celery_app, memory_exporter):
    @celery_app.task
    def fn_exception():
        raise Exception("Task class is failing")

    result = fn_exception.apply()

    assert result.failed() is True
    assert "Task class is failing" in result.traceback

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is False
    assert span.name == "run/test_celery_functional.fn_exception"
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "FAILURE"
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_exception"
    )
    assert span.status.status_code == StatusCode.ERROR
    assert len(span.events) == 1
    event = span.events[0]
    assert event.name == "exception"
    assert event.attributes[SpanAttributes.EXCEPTION_TYPE] == "Exception"
    assert SpanAttributes.EXCEPTION_MESSAGE in event.attributes
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert "Task class is failing" in span.status.description


def test_fn_exception_expected(celery_app, memory_exporter):
    @celery_app.task(throws=(MyException,))
    def fn_exception():
        raise MyException("Task class is failing")

    result = fn_exception.apply()

    assert result.failed() is True
    assert "Task class is failing" in result.traceback

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is False
    assert span.status.status_code == StatusCode.ERROR
    assert span.name == "run/test_celery_functional.fn_exception"
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "FAILURE"
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_exception"
    )
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )


def test_fn_retry_exception(celery_app, memory_exporter):
    @celery_app.task
    def fn_exception():
        raise Retry("Task class is being retried")

    result = fn_exception.apply()

    assert result.failed() is False
    assert "Task class is being retried" in result.traceback

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is True
    assert span.status.status_code == StatusCode.UNSET
    assert span.name == "run/test_celery_functional.fn_exception"
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "RETRY"
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.fn_exception"
    )
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )


def test_class_task(celery_app, memory_exporter):
    class BaseTask(celery_app.Task):
        def run(self):
            return 42

    task = BaseTask()
    # register the Task class if it's available (required in Celery 4.0+)
    register_task = getattr(celery_app, "register_task", None)
    if register_task is not None:
        register_task(task)

    result = task.apply()

    assert result.successful() is True
    assert result.result == 42

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is True
    assert span.name == "run/test_celery_functional.BaseTask"
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.BaseTask"
    )
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "SUCCESS"
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )


def test_class_task_exception(celery_app, memory_exporter):
    class BaseTask(celery_app.Task):
        def run(self):
            raise Exception("Task class is failing")

    task = BaseTask()
    # register the Task class if it's available (required in Celery 4.0+)
    register_task = getattr(celery_app, "register_task", None)
    if register_task is not None:
        register_task(task)

    result = task.apply()

    assert result.failed() is True
    assert "Task class is failing" in result.traceback

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is False
    assert span.name == "run/test_celery_functional.BaseTask"
    assert (
        span.attributes.get("celery.task_name")
        == "test_celery_functional.BaseTask"
    )
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "FAILURE"
    assert span.status.status_code == StatusCode.ERROR
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )
    assert "Task class is failing" in span.status.description


def test_class_task_exception_expected(celery_app, memory_exporter):
    class BaseTask(celery_app.Task):
        throws = (MyException,)

        def run(self):
            raise MyException("Task class is failing")

    task = BaseTask()
    # register the Task class if it's available (required in Celery 4.0+)
    register_task = getattr(celery_app, "register_task", None)
    if register_task is not None:
        register_task(task)

    result = task.apply()

    assert result.failed() is True
    assert "Task class is failing" in result.traceback

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is False
    assert span.status.status_code == StatusCode.ERROR
    assert span.name == "run/test_celery_functional.BaseTask"
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "FAILURE"
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )


def test_shared_task(celery_app, memory_exporter):
    """Ensure Django Shared Task are supported"""

    @celery.shared_task
    def add(x, y):
        return x + y

    result = add.apply([2, 2])
    assert result.result == 4

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1

    span = spans[0]

    assert span.status.is_ok is True
    assert span.name == "run/test_celery_functional.add"
    assert (
        span.attributes.get("celery.task_name") == "test_celery_functional.add"
    )
    assert span.attributes.get("celery.action") == "run"
    assert span.attributes.get("celery.state") == "SUCCESS"
    assert (
        span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )


@mark.skip(reason="inconsistent test results")
def test_apply_async_previous_style_tasks(
    celery_app, celery_worker, memory_exporter
):
    """Ensures apply_async is properly patched if Celery 1.0 style tasks are
    used even in newer versions. This should extend support to previous versions
    of Celery."""

    class CelerySuperClass(celery.task.Task):
        abstract = True

        @classmethod
        def apply_async(cls, args=None, kwargs=None, **kwargs_):
            return super().apply_async(args=args, kwargs=kwargs, **kwargs_)

        def run(self, *args, **kwargs):
            if "stop" in kwargs:
                # avoid call loop
                return
            CelerySubClass.apply_async(args=[], kwargs={"stop": True}).get(
                timeout=ASYNC_GET_TIMEOUT
            )

    class CelerySubClass(CelerySuperClass):
        pass

    celery_worker.reload()

    task = CelerySubClass()
    result = task.apply()

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 3

    async_span, async_run_span, run_span = spans

    assert run_span.status.is_ok is True
    assert run_span.name == "run/test_celery_functional.CelerySubClass"
    assert (
        run_span.attributes.get("celery.task_name")
        == "test_celery_functional.CelerySubClass"
    )
    assert run_span.attributes.get("celery.action") == "run"
    assert run_span.attributes.get("celery.state") == "SUCCESS"
    assert (
        run_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        == result.task_id
    )

    assert async_run_span.status.is_ok is True
    assert async_run_span.name == "run/test_celery_functional.CelerySubClass"
    assert (
        async_run_span.attributes.get("celery.task_name")
        == "test_celery_functional.CelerySubClass"
    )
    assert async_run_span.attributes.get("celery.action") == "run"
    assert async_run_span.attributes.get("celery.state") == "SUCCESS"
    assert (
        async_run_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        != result.task_id
    )

    assert async_span.status.is_ok is True
    assert (
        async_span.name == "apply_async/test_celery_functional.CelerySubClass"
    )
    assert (
        async_span.attributes.get("celery.task_name")
        == "test_celery_functional.CelerySubClass"
    )
    assert async_span.attributes.get("celery.action") == "apply_async"
    assert (
        async_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)
        != result.task_id
    )
    assert async_span.attributes.get(
        SpanAttributes.MESSAGING_MESSAGE_ID
    ) == async_run_span.attributes.get(SpanAttributes.MESSAGING_MESSAGE_ID)


def test_custom_tracer_provider(celery_app, memory_exporter):
    @celery_app.task
    def fn_task():
        return 42

    resource = resources.Resource.create({})
    tracer_provider = TracerProvider(resource=resource)
    span_processor = export.SimpleSpanProcessor(memory_exporter)
    tracer_provider.add_span_processor(span_processor)

    trace_api.set_tracer_provider(tracer_provider)

    CeleryInstrumentor().uninstrument()
    CeleryInstrumentor().instrument(tracer_provider=tracer_provider)

    fn_task.apply()

    spans_list = memory_exporter.get_finished_spans()
    assert len(spans_list) == 1

    span = spans_list[0]
    assert span.resource == resource



import logging
import requests

logger = logging.getLogger(__name__)
'''
获取agent 下载地址。
'''


def get_agent_path(env):
    if env == "dev" :
        return "http://10.50.208.198:9624/dy/api/paas-mgr/nationarmsadmin/cmparms/app/agent/download?agentName=python_agent"
    
    return "https://arms.ctyun.cn/dy/api/paas-mgr/nationarmsadmin/cmparms/app/agent/download?agentName=python_agent"
    

'''
判断网络是否通
'''


def check_network(url, timeout=1):
    try:
        response = requests.request("GET", url, timeout=timeout)
        if response.status_code == 200:
            return True
    except Exception:
        return False
